/**
 *
 * Előadás különböző kódrészletei
 *
 */

#include <iostream>
#include <cstdlib>
#include <cstring>

using std::cout;
using std::endl;

template <typename T, class S>
T leg1(const T a[], int n) {
  T tmp = a[0];
  for (int i = 1; i < n; ++i)
    if (S::select(a[i], tmp)) tmp = a[i];
  return tmp;
}

struct kisebb_int {
  static bool select(const int a, const int b) { return a < b; }
};

// Lehet egy sablonból generálódó osztály is
template<typename T>
struct nagyobb {	// szokásos nagyobb reláció
  static bool select(const T a, const T b) { return a > b; }
};

template <typename T, bool select(T, T)>
T leg2(const T a[], int n) {
  T tmp = a[0];
  for (int i = 1; i < n; ++i)
    if (select(a[i], tmp)) tmp = a[i];
  return tmp;
}

template<class T>
bool kisebbFv (T a, T b) {  return a < b; }

template<class T>
struct kisebbObj {
    bool operator()(const T& a, const T& b) {
		return a < b;
    }
};

template <typename T, typename S>
T legElem(const T a[], int n, S sel) {
    T tmp = a[0];
    for (int i = 1; i < n; ++i)
        if (sel(a[i], tmp)) tmp = a[i];
    return tmp;
}

template <typename T, class F>
void rendez (T a[], int n, F cmp) {
  for (int i = 1; i < n; i++) {
    T ujelem = a[i]; int j = i-1;
      while (j >= 0 && cmp(ujelem, a[j])) {
        a[j+1] = a[j]; j--;
      }
      a[j+1] = ujelem;
  }
}

template <typename T>
struct kisebb {
    bool operator()(T a, T b) { return a < b; }
};

template <typename T>
void Print(const T t[], int n) {
    for (int i = 0; i < n-1; ++i)
        cout << t[i] << ",";
    cout << t[n-1] << endl;
}

template <class T> struct Pont   { T x, y; };
template <class T> struct Pont3D { T x, y, z; };


template <class S,
	    template <class T> class P = Pont >
struct Idom {
    P<S> origo;  // idom origója
};

template <class T1, typename T2, int i = 0>
struct V {
    T1 a1; T2 a2; int x;
    V() { x = i;}
};

/// Konstruktor specializációja <int, int>-re
template<>
V<int, int>::V() { a1=a2=x=0; }

/// Template mint paraméter
template <class T> struct miez { T a0; };
template <class K, template <class T> class C = miez>
struct S {
    C<K> a1;
};

// paraméterként kapott típus és konstansa, ami
// csak integrális, vagy pointer típus lehet
template <typename T, T c = T()>
struct V1 {
    T val;
    V1() { val = c;}
};

/// Faktorialis
template <int N>
struct fakt {
    enum  { fval =  N * fakt<N-1>::fval };
};

template <>
struct fakt<0> {
    enum  { fval =  1 };
};


int main() {
// Használat:
    int it[9] = {-5, -4, -3, -2, -1, 0, 1, 2, 3 };
    double dt[5] =  { .0,  .1,  .2,  .3,  4.4 };

    cout << "Legelem:" << endl;
    cout << leg1<int, kisebb_int>(it, 9) << endl;  		// -5
    cout << leg1<int, nagyobb<int>  >(it, 9) << endl; 	// 3
    cout << leg1<double, nagyobb<double>  >(dt, 5);	// 4.4
    cout << leg2<int, kisebbFv<int> >(it, 9);
    cout << legElem<int, kisebbObj<int> >(it, 9, kisebbObj<int>() );

    cout << "Rendez:" << endl;
    rendez(it, 9, kisebb<int>());
    Print(it, 9);

    Idom<int> sikidom1;     sikidom1.origo.x = 10;
    Idom<double> sikidom2;  sikidom2.origo.x = 4.5;
    Idom<int, Pont3D> test; test.origo.z = 3;

    V<int, char, 4> v1;
    V<double, int, 8> v2;
    V<int, int> v3;     /// specializált példány

    cout << "v1.x: " << v1.x << endl;
    cout << "v2.x: " << v2.x << endl;
    cout << "v3.x: " << v3.x << endl;

    V1<int, 30> v30;
    V1<int*> v0;
    cout << "v30.val: " << v30.val << endl;
    cout << "v0.val: " << v0.val << endl;

    cout << "fakt(5): " << fakt<5>::fval << endl;
}


